"""Defines error cases which can occur using Axii."""

import sys
import traceback
import typing as ty


def handle_exception(expected_because: ty.Optional[str] = None):
    from armarx_setup.core.config import config as axii_config

    console = axii_config.console
    dev_mode = axii_config.get("dev_mode")
    print_exception_traceback = axii_config.get("print_exception_traceback")
    if dev_mode and print_exception_traceback:
        if expected_because is not None:
            exception_class, _, exception_traceback = sys.exc_info()
            exc_type = exception_class.__name__
            last_trace = traceback.extract_tb(exception_traceback)[-1]
            location = f"'{last_trace.filename}':{last_trace.lineno} ('{last_trace.name}')"
            message = f"{expected_because}" + "." if not expected_because.endswith(".") else ""

            console.print(f"Handled '{exc_type}' in {location}. {message}")
        else:
            console.print_exception()
    # TODO: Log errors.


def print_exception_message(exception: Exception):
    from armarx_setup.core.config import config as axii_config

    type_indicator = ""
    if not isinstance(exception, AxiiError):
        type_indicator = exception.__class__.__name__

    console = axii_config.console
    message = f"{type_indicator}{exception}"
    if not message.endswith("."):
        message += "."
    console.print_error(f"Error:  {message}")


AXII_ERROR = 10
WORKSPACE_ERROR = 50
MODULE_ERROR = 90

# Special error codes.
UNCLASSIFIED_ERROR = 1
SUBPROCESS_FAILED_ERROR = 252
USER_ABORTED = 253
NOT_YET_IMPLEMENTED_ERROR = 254


class AxiiError(Exception):
    """Generic base exception for Axii."""

    def __init__(self, message: str, axii_return_code: int = UNCLASSIFIED_ERROR):
        message = f"{message}." if not message.endswith(".") else message
        super().__init__(message)
        self.axii_return_code = axii_return_code


class AxiiSpecificError(AxiiError):
    """Describes an Axii-specific error."""

    def __init__(self, message, error_number):
        assert 0 <= error_number <= 39, f"Error number {error_number} out of range for scope " \
                                        f"'Axii'."
        super().__init__(message, axii_return_code=AXII_ERROR + error_number)


class WorkspaceSpecificError(AxiiError):
    """Describes a workspace-related error."""

    def __init__(self, message, error_number):
        assert 0 <= error_number <= 39, f"Error number {error_number} out of range for scope " \
                                        f"'workspace'."
        super().__init__(message, axii_return_code=WORKSPACE_ERROR + error_number)


class ModuleSpecificError(AxiiError):
    """Describes a module-related error."""

    def __init__(self, message, error_number):
        assert 0 <= error_number <= 39, f"Error number {error_number} out of range for scope " \
                                        f"'module'."
        super().__init__(message, axii_return_code=MODULE_ERROR + error_number)


class PromptsDisabled(AxiiSpecificError):
    """Describes that prompts are disabled but Axii has insufficient information to proceed."""

    def __init__(self):
        super().__init__("Prompts are disabled.", error_number=30)


class ShellIntegrationNotInstalled(AxiiSpecificError):
    """Describes that Axii's shell integration was not installed properly."""

    def __init__(self):
        super().__init__("Axii is not installed properly.", error_number=31)


class UnsupportedShell(AxiiSpecificError):
    """Describes that the given shell is unsupported."""

    def __init__(self, path):
        super().__init__(f"Unsupported shell in $SHELL='{path}'.", error_number=32)


class NoShellRunComFile(AxiiSpecificError):
    """Describes that no run com (.bashrc, .zshrc, ...) file for the given shell was found."""

    def __init__(self, shell, run_com_file):
        super().__init__(f"No shell run com (rc) file found for {shell} in '{run_com_file}'.",
                         error_number=33)


class NoSuchConfigVariable(AxiiSpecificError):
    """Describes that no such configuration variable with the given name exists."""

    def __init__(self, variable_name):
        super().__init__(f"No such config variable '{variable_name}'.", error_number=10)

        self.variable_name = variable_name


class InvalidConfigurationValue(AxiiSpecificError):
    """Describes that the value of the given configuration variable is invalid."""

    def __init__(self, variable, value, description):
        if not description.endswith("."):
            description = description + "."
        super().__init__(
            f"Invalid value '{value}' of variable {variable}. \n"
            f"{description}",
            error_number=11
        )


class NoWorkspaceActive(WorkspaceSpecificError):
    """Describes that no workspace is active even though it was expected."""

    def __init__(self):
        super().__init__("No workspace active.", error_number=1)


class NoSuchWorkspace(WorkspaceSpecificError):
    """Describes that no workspace with the given name exists."""

    def __init__(self, workspace_name):
        super().__init__(f"No such workspace '{workspace_name}'.", error_number=2)

        self.workspace_name = workspace_name


class WorkspaceAlreadyExists(WorkspaceSpecificError):
    """Describes that a workspace with the given name already exists."""

    def __init__(self, name):
        super().__init__(f"Workspace '{name}' already exists.", error_number=3)

        self.name = name


class WorkspaceWasBuiltWithOtherOS(WorkspaceSpecificError):
    """Describes that a workspace was previously built with another OS or OS version."""

    def __init__(self, workspace_os_fingerprint, actual_os_fingerprint):
        super().__init__(f"Workspace was built with another OS or OS version, which is not supported.  The OS that "
                         f"built the workspace is identified by '{workspace_os_fingerprint}', this workspace is "
                         f"identified by '{actual_os_fingerprint}'.\n"
                         f"You can use 'axii workspace clean' to remove all build files "
                         f"and make the workspace compatible with the current OS.",
                         error_number=4)


class ModuleInfoNotFound(ModuleSpecificError):
    """Describes that a module info for the given module could not be found."""

    def __init__(self, name: str, tried_paths: ty.List[str], similar_modules=None,
                 database_module=None):
        super().__init__(
            f"Module information for '{name}' not found. Tried: "
            + "".join(f"\n  {path}" for path in tried_paths),
            error_number=1
        )

        self.name = name
        self.similar_modules = similar_modules
        self.database_module = database_module


class ModuleAlreadyExists(ModuleSpecificError):
    """Describes that a module with the given name already exists."""

    def __init__(self, name: str):
        super().__init__(f"Module '{name}' already exists.", error_number=2)

        self.name = name


class NoSuchModule(ModuleSpecificError):
    """Describes that no such module with the given name exists."""

    def __init__(self, name: str, similar_modules=None):
        super().__init__(f"There is no module with name '{name}.'", error_number=3)

        self.name = name
        self.similar_modules = similar_modules


class CyclicDependencies(ModuleSpecificError):
    """Describes that modules in the module graph show a cyclic dependency."""

    def __init__(self, caused_by: "nx.NetworkXUnfeasible"):
        super().__init__(caused_by, error_number=4)


class ModuleRequiresItself(ModuleSpecificError):
    """Describes that a module depends on itself."""

    def __init__(self, module_name: str):
        super().__init__(
            f"The module '{module_name}' requires itself. \n"
            f"Remove the entry '{module_name}' from the 'required modules' "
            f"of the definition of '{module_name}' and try again.",
            error_number=5
        )


class MalformedAuthorReference(ModuleSpecificError):
    """Describes that the author information of a module is malformed."""

    def __init__(self, author_reference):
        super().__init__(
            f"Given author reference '{author_reference}' is malformed. Expected "
            f"'Display Name <email@example.com>' style reference.",
            error_number=10
        )


class StepFailed(ModuleSpecificError):
    """Describes that the execution of a step failed."""

    def __init__(self, module_name, step, step_result):
        summary = step_result.summary
        if not summary.startswith("."):
            summary += "."
        super().__init__(
            f"Failed to execute step '{step}' on module '{module_name}':  {step_result.summary}",
            error_number=20
        )
        self.module_name = module_name
        self.step = step
        self.step_result = step_result


class NoSuchCommandCwd(AxiiError):
    """Describes that a command could not be executed."""

    def __init__(self, cmd: ty.List[str], cwd: str):
        self.cmd = cmd
        self.cwd = cwd
        super().__init__(f"Could not execute command, cwd '{cwd}' does not exist.",
                         axii_return_code=SUBPROCESS_FAILED_ERROR)


class CommandFailed(AxiiError):
    """Describes that a command was executed but exited with a non-zero exit code."""

    def __init__(
            self,
            cmd: ty.List[str],
            cwd: ty.Optional[str] = None,
            stdout: ty.Optional[str] = None,
            stderr: ty.Optional[str] = None,
            return_code: ty.Optional[int] = None,
            summary: ty.Optional[str] = None
    ):
        self.pretty_cmd: str
        if isinstance(cmd, str):
            self.pretty_cmd = cmd
        else:
            self.pretty_cmd = " ".join(cmd)
        in_cwd = "" if cwd is None else f"in '{cwd}' "
        message = f"Running '{self.pretty_cmd}' {in_cwd}failed with return code {return_code}."
        super().__init__(message, axii_return_code=SUBPROCESS_FAILED_ERROR)

        self.cmd = cmd
        self.cwd = cwd
        self.stdout = stdout
        self.stderr = stderr
        self.return_code = return_code
        self.summary = summary


class UserAborted(AxiiError):
    """Describes that an action was aborted by the user."""

    def __init__(self):
        super().__init__("Aborted by user.", axii_return_code=USER_ABORTED)


class NotYetImplemented(AxiiError):
    """Describes that a given feature has not yet been implemented."""

    def __init__(self, feature):
        super().__init__(f"This feature ({feature}) is not yet implemented. Search for open issues "
                         f"or open a new one to raise awareness.",
                         axii_return_code=NOT_YET_IMPLEMENTED_ERROR)

        self.feature = feature
