"""Supplies strategies for the `build` step."""

import dataclasses as dc
import os
from typing import Dict, List, Optional

from armarx_setup.core import error
from armarx_setup.core.config import config as axii_config
from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook


def clear_directory(path: str, hidden=True, exceptions=None):
    import glob, shutil
    assert os.path.isdir(path)

    if exceptions is None:
        exceptions: List[str] = []

    entries = glob.glob(os.path.join(path, "*"))
    if hidden:
        entries += glob.glob(os.path.join(path, ".*"))

    for exception in exceptions:
        exception_path = os.path.join(path, exception)
        if exception_path in entries:
            entries.remove(exception_path)

    for entry in entries:
        if os.path.isfile(entry) or os.path.islink(entry):
            os.unlink(entry)
        elif os.path.isdir(entry):
            shutil.rmtree(entry)


@dc.dataclass
class CMake(Hook):
    env: Dict[str, str] = dc.field(default_factory=dict)

    def generate_requirements(self) -> List[str]:
        return common.CMAKE_DEPENDENCIES

    def __call__(self, *args, **kwargs) -> common.StepResult:
        job_number = axii_config.get('build_job_number')
        give_up_building = False  # We give up once we repeatedly failed to build the project.

        step_results = []
        last_step_result = None

        while not give_up_building:
            build_dir = self.ctx.build_dir
            if not os.path.isdir(build_dir):
                raise error.AxiiError(f"Build directory '{build_dir}' does not exist.")

            try:
                env = self.ctx.ws_config.global_.build.cmake.env.copy()
            except AttributeError:
                env = {}
            env = {"CMAKE_BUILD_PARALLEL_LEVEL": f"{job_number}",
                   **env, **self.ctx.env, **self.env}

            cmake = "cmake --build . --target all".split()

            last_step_result = common.execute_command(command=cmake, cwd=build_dir, env=env)
            step_results.append(last_step_result)

            # This error happens if we run out of memory while building and any OOM kills the build process.
            out_of_memory_indicator = "signal terminated program cc1plus"
            if last_step_result.status == common.StepResultStatus.FAILURE \
                    and (out_of_memory_indicator in last_step_result.error_output or \
                    out_of_memory_indicator in last_step_result.output):  # Ninja prints fatal errors on stdout...
                console = axii_config.console
                console.print_error("Axii detected that the compilation process was terminated during compilation, "
                                    "most likely due to the system running out of memory.")

                if job_number > 1:
                    # Half job number and re-try.
                    new_job_number = job_number // 2
                    console.print(f"Retrying build with {new_job_number} instead of {job_number} jobs.")
                    console.print()
                    job_number = new_job_number
                else:
                    # Give up.
                    console.print_error("Giving up after trying to build with one job.")
                    give_up_building = True
            else:
                give_up_building = True  # If we failed even though we only had one job, we have to give up.

        assert len(step_results) > 0, "No command was executed in build phase."

        step_result = common.StepResult.from_sub_steps_results(step_results)

        if step_result.status == common.StepResultStatus.FAILURE:
            if last_step_result.status == common.StepResultStatus.SUCCESS:
                # StepResultStatus.from_sub_steps_results will set the status of the combined step result to FAILURE if
                # one of the sub-steps ended in FAILURE.  However, if the status of last_step_result was successful,
                # then the whole run was successful (with warnings though) and we can set it as such.
                step_result.status = common.StepResultStatus.SUCCESS_WITH_WARNING
                step_result.summary = "Several tries for building were necessary"
                return step_result
            else:
                if len(step_results) > 1:
                    step_result.summary = "Could not build the project after several tries"
                return step_result

        no_op_messages = [
            "make: Nothing to be done for",
            "ninja: no work to do."
        ]

        if any(msg in step_result.output for msg in no_op_messages):
            step_result.status = common.StepResultStatus.SKIPPED

        return step_result

    def clean(self):
        build_dir = self.ctx.build_dir
        if os.path.isdir(build_dir):
            clear_directory(build_dir, exceptions=[".gitkeep", ".gitignore"])


@dc.dataclass
class Run(common.Run):

    def clean(self):
        pass

    def __call__(self, *args, **kwargs) -> common.StepResult:
        job_number = axii_config.get('build_job_number')

        env = {"BUILD_JOB_NUMBER": f"{job_number}",
               **self.ctx.env}

        return super().run(env=env)


@dc.dataclass
class Build(Hook):
    cmake: Optional[CMake] = None
    run: Optional[Run] = None

    def __post_init__(self):
        from armarx_setup.core.util.dataclass import sanitize_type

        kwargs = { "ctx": self.ctx }
        sub_step_enabled_kwargs = { "list_factory": common.SubStepAdapter.list_factory, **kwargs }

        self.cmake = sanitize_type(self.cmake, CMake, **sub_step_enabled_kwargs)
        self.run = sanitize_type(self.run, Run, **sub_step_enabled_kwargs)

    def _get_sub_hooks(self) -> List[Hook]:
        return [self.cmake, self.run]

    def clean(self):
        for option in self._get_sub_hooks():
            if option is not None:
                option.clean()
