"""Supplies common strategies for the several steps."""

import dataclasses as dc
import enum
import os
import typing as ty

from armarx_setup.core import error
from armarx_setup.core.util import environ_context
from armarx_setup.core.module.context import Hook, Context
from armarx_setup.core.util import dataclass

CMAKE_DEPENDENCIES = ["tools/default_cmake"]


def execute_command(
        command: ty.List[str],
        cwd: str,
        env: ty.Optional[ty.Dict[str, str]] = None,
        success_summary: ty.Optional[str] = None,
        failure_summary: ty.Optional[str] = None,
        executor: ty.Optional[ty.Callable] = None
):
    if executor is None:
        from armarx_setup.core.config import config as axii_config
        executor = axii_config.executor
    execute = executor

    command_str = " ".join(command)

    if success_summary is None:
        success_summary = "Command succeeded"
    if failure_summary is None:
        failure_summary = "Command exited with return code {return_code}"

    stderr: str = ""

    def capture_stderr(text: str):
        nonlocal stderr
        stderr += text + "\n"

    try:
        stdout = execute(command, cwd=cwd, env=env, stderr_callback=capture_stderr)
    except error.CommandFailed as e:
        status = StepResultStatus.FAILURE
        summary = failure_summary.format(return_code=e.return_code)
        stdout = e.stdout
        stderr = e.stderr
    else:
        status = StepResultStatus.SUCCESS
        summary = success_summary

    return StepResult(status=status, summary=summary, executed_command=command_str,
                      output=stdout, error_output=stderr)


def resolve_path(path: str, ctx: Context) -> str:
    with environ_context.EnvironContext(ctx.env):
        path = os.path.expanduser(os.path.expandvars(path))

    if os.path.isabs(path):
        absolute_path = path
    else:
        info_dir = os.path.dirname(ctx.module_info_path)
        absolute_path = os.path.join(info_dir, path)

    return absolute_path


class StepResultStatus(enum.Enum):
    SUCCESS = enum.auto()
    FAILURE = enum.auto()
    SUCCESS_WITH_WARNING = enum.auto()
    SKIPPED = enum.auto()


@dc.dataclass
class StepResult:
    status: StepResultStatus
    summary: str
    executed_command: ty.Optional[str] = None
    output: ty.Optional[str] = None
    error_output: ty.Optional[str] = None

    sub_steps: ty.List["StepResult"] = dc.field(default_factory=list)

    @staticmethod
    def from_sub_steps_results(sub_steps_results):
        if len(sub_steps_results) == 1:
            return sub_steps_results[0]

        statuses = [s.status for s in sub_steps_results]
        if StepResultStatus.FAILURE in statuses:
            status = StepResultStatus.FAILURE
            info_text = [s.summary for s in sub_steps_results
                         if s.status == StepResultStatus.FAILURE]
            summary = f"Some executed sub steps failed: {', '.join(info_text)}"
        elif StepResultStatus.SUCCESS_WITH_WARNING in statuses:
            status = StepResultStatus.SUCCESS_WITH_WARNING
            info_text = [s.summary for s in sub_steps_results
                         if s.status == StepResultStatus.SUCCESS_WITH_WARNING]
            summary = f"Some executed sub steps contained warnings: {', '.join(info_text)}"
        elif StepResultStatus.SUCCESS in statuses:
            status = StepResultStatus.SUCCESS
            summary = "All executed sub steps succeeded"
        else:
            status = StepResultStatus.SKIPPED
            summary = "Sub steps have been skipped"

        outputs = [s.output for s in sub_steps_results if s.output is not None]
        error_outputs = [s.error_output for s in sub_steps_results if s.error_output is not None]

        return StepResult(status=status, summary=summary, sub_steps=sub_steps_results,
                          output="\n".join(outputs), error_output="\n".join(error_outputs))


@dc.dataclass
class SubStepAdapter(Hook):
    sub_steps: ty.List[Hook]
    step_type: type

    def __post_init__(self):
        kwargs = { "ctx": self.ctx }

        sub_steps = []
        for sub_step in self.sub_steps:
            sub_steps.append(dataclass.sanitize_type(sub_step, self.step_type, **kwargs))
        self.sub_steps = sub_steps

    def __call__(self, *args, **kwargs) -> StepResult:
        sub_steps_results = []
        for sub_step in self.sub_steps:
            sub_steps_results.append(sub_step(*args, **kwargs))
        if len(sub_steps_results) == 1:
            return sub_steps_results[0]
        return StepResult.from_sub_steps_results(sub_steps_results=sub_steps_results)

    @classmethod
    def list_factory(cls, val, step_cls, **kwargs):
        return cls(sub_steps=val, step_type=step_cls, **kwargs)

    def _get_sub_hooks(self) -> ty.List["Hook"]:
        return self.sub_steps


@dc.dataclass
class Run(Hook):
    comment: str = ""
    command: ty.Optional[ty.Union[ty.List[str], str]] = None
    run_in_shell: bool = False

    def __post_init__(self):
        import shlex

        if isinstance(self.command, str):
            self.command = shlex.split(self.command)

        if self.run_in_shell:
            self.command = ["bash", "-c", " ".join(self.command)]

    def __call__(self, *args, **kwargs) -> StepResult:
        return self.run(*args, **kwargs)

    def run(self, env=None, *args, **kwargs) -> StepResult:
        from armarx_setup.core.config import config as axii_config

        console = axii_config.console

        if env is None:
            env = self.ctx.env

        os.makedirs(self.ctx.build_dir, exist_ok=True)
        cwd = self.ctx.build_dir

        if self.comment != "":
            console.print()
            console.print(self.comment)

        with environ_context.EnvironContext(env=self.ctx.env):
            self.command = [os.path.expandvars(p) for p in self.command]

        return execute_command(self.command, cwd=cwd, env=env)


class Source(Hook):

    def __init__(self, paths: ty.List[str], ctx=None):
        super().__init__(ctx=ctx)

        self.paths: ty.List[str] = paths
        for path in self.paths:
            if not isinstance(path, str):
                raise error.AxiiError(f"Invalid path entry: '{path}'")

    def generate_env(self, env: "EnvFile"):
        by = "global variables"
        if self.ctx.name:
            by = self.ctx.name
        for path in self.paths:
            path = resolve_path(path, ctx=self.ctx)
            env.add_source(path=path, by=by, write=False)


if __name__ == '__main__':
    from armarx_setup.core.workspace.env import EnvFile
