"""Supplies strategies for the `install` step."""

import dataclasses as dc
import os
from typing import Dict, List, Optional

from armarx_setup.core import error
from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook
from armarx_setup.core.util import commands, environ_context as env_ctx
from armarx_setup.core.module.common import Run


@dc.dataclass
class CMake(Hook):
    dest_dir: Optional[str] = None

    def generate_requirements(self) -> List[str]:
        return common.CMAKE_DEPENDENCIES

    def __call__(self, *args, **kwargs):
        from armarx_setup.core.config import config as axii_config

        build_dir = self.ctx.build_dir
        if not os.path.isdir(build_dir):
            raise IOError(f"Build directory '{build_dir}' does not exist.")

        env = {**self.ctx.env}

        if self.dest_dir is not None:
            axii_config.console.print(f"Property 'dest_dir' of module '{self.ctx.module_name}' is "
                                      f"deprecated.  Use 'CMAKE_INSTALL_PREFIX'.")
            env["DESTDIR"] = self.dest_dir

        cmake = "cmake --install .".split()

        return common.execute_command(cmake, cwd=build_dir, env=env)


class Env(Hook):

    def __init__(self, ctx=None, **kwargs):
        super().__init__(ctx=ctx)

        self.env: Dict[str, str] = kwargs
        for k in self.env:
            v = self.env[k]

            if not (isinstance(k, str) and isinstance(v, str)):
                raise error.AxiiError(f"Invalid env entry: '{k}': '{v}'")

            self.env[k] = v

    def generate_env(self, env: "EnvFile"):
        for k, v in self.env.items():
            if self.ctx.path:
                env_vars = self.ctx.env
                for key, val in env_vars.items():
                    v = v.replace(f"${key}", val)
            env.add(k, v, by=self.ctx.name)

    def __repr__(self):
        return ("Env({"
                + ", ".join(f"'{k}': '{v}'" for k, v in self.env.items())
                + "})")


class Executables(Hook):

    def __init__(self, ctx=None, **kwargs):
        super().__init__(ctx=ctx)

        self.execs = kwargs

    def __call__(self, *args, **kwargs) -> common.StepResult:

        assert "home" in kwargs, "Workspace home was not given as parameter"

        home = kwargs["home"]

        for exec_name, exec_config in self.execs.items():
            kill_trap = exec_config["kill_trap"] if "kill_trap" in exec_config else False
            fork = "&" if kill_trap else ""

            with env_ctx.EnvironContext(self.ctx.env, verbose=self.ctx.verbose):
                home.install_executable(exec_name, exec_config, kill_trap, fork)

        return common.StepResult(status=common.StepResultStatus.SUCCESS,
                                 summary="Created executable")


@dc.dataclass
class Install(Hook):
    env: Optional[Env] = None
    cmake: Optional[CMake] = None
    source: Optional[common.Source] = None
    run: Optional[Run] = None
    executables: Optional[Executables] = None

    def __post_init__(self):
        from armarx_setup.core.util.dataclass import sanitize_type

        kwargs = { "ctx": self.ctx }
        sub_step_enabled_kwargs = { "list_factory": common.SubStepAdapter.list_factory, **kwargs }

        self.env = sanitize_type(self.env, Env, **kwargs)
        self.cmake = sanitize_type(self.cmake, CMake, **sub_step_enabled_kwargs)
        self.source = sanitize_type(self.source, common.Source, **kwargs)
        self.run = sanitize_type(self.run, common.Run, **sub_step_enabled_kwargs)
        self.executables = sanitize_type(self.executables, Executables, **kwargs)

    def _get_sub_hooks(self) -> List["Hook"]:
        return [self.env, self.cmake, self.source, self.run, self.executables]

    def is_step_empty(self):
        sub_steps = [self.cmake, self.source, self.run, self.executables]  # Env is a separate step.
        return all(s is None for s in sub_steps)


if __name__ == '__main__':
    from armarx_setup.core.workspace.env import EnvFile
