"""Supplies strategies for the `update` step."""

from collections import OrderedDict
import dataclasses as dc
import os
from typing import List, Optional
import requests

from armarx_setup.core import error
from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook
from armarx_setup.core.module.update import Git, Download
from armarx_setup.core.util import commands, environ_context


@dc.dataclass
class Archive(Hook):
    @dc.dataclass
    class Extract:
        type: Optional[str] = None
        """Directory created by extraction."""
        created_directory: Optional[str] = None
        extract_into: Optional[str] = None

    url: Optional[str] = None
    extract: Optional[Extract] = None
    make_executable: bool = False

    def __post_init__(self):
        if isinstance(self.extract, bool):
            self.extract = self.Extract()
        else:
            from armarx_setup.core.util import dataclass
            self.extract = dataclass.sanitize_type(self.extract, self.Extract)

    @property
    def expanded_url(self):
        if self.url is not None:
            with environ_context.EnvironContext(self.ctx.env):
                return os.path.expandvars(self.url)
        return None

    @property
    def expanded_extract_into(self):
        if self.extract is not None and self.extract.extract_into is not None:
            with environ_context.EnvironContext(self.ctx.env):
                return os.path.expandvars(self.extract.extract_into)
        return None

    @property
    def archive_filename(self) -> str:
        return os.path.basename(self.expanded_url)

    @property
    def archive_filepath(self) -> str:
        return os.path.join(self.ctx.path, self.archive_filename)

    def __call__(self, *args, **kwargs) -> common.StepResult:
        os.makedirs(self.ctx.path, exist_ok=True)

        status = None
        summary = None
        actions = []

        refresh_extract = False
        if not os.path.isfile(self.archive_filepath):
            step_result = self.download()
            if step_result.status == common.StepResultStatus.FAILURE:
                return step_result
            actions.append("downloaded")
            refresh_extract = True

        if self.extract is not None:
            try:
                done = self.do_extract(refresh=refresh_extract)
            except error.AxiiError as e:
                status = common.StepResultStatus.FAILURE
                summary = "Failed extracting"
            else:
                status = common.StepResultStatus.SUCCESS
                if done:
                    actions.append("extracted")

        if self.make_executable and not os.access(self.archive_filepath, os.X_OK):
            os.chmod(self.archive_filepath, 0o750)
            actions.append("made executable")

        if summary is None:
            if actions:
                summary = f"{(' and '.join(actions)).capitalize()} '{self.archive_filename}'."
            else:
                status = common.StepResultStatus.SKIPPED
                summary = "Nothing to do"

        return common.StepResult(status, summary=summary, output=summary)

    def download(self):
        response = requests.get(self.expanded_url)
        if response.status_code >= 300:
            return common.StepResult(common.StepResultStatus.FAILURE,
                                     summary="Failed downloading file",
                                     error_output=f"Downloading failed with status code "
                                                  f"{response.status_code}: {response.reason}")

        with open(self.archive_filepath, 'wb') as file:
            file.write(response.content)
        assert os.path.isfile(self.archive_filepath)
        return common.StepResult(common.StepResultStatus.SUCCESS,
                                 summary="Downloaded file",
                                 output="Successfully downloaded file")

    def do_extract(self, refresh: bool):
        extractors = {
            ".tar.gz": self.extract_tar,
            ".tar.xz": self.extract_tar,
            ".tgz": self.extract_tar,
            ".zip": self.extract_zip,
            ".deb": self.extract_deb
        }

        for ext, fn in extractors.items():
            if self.matches_extension(ext):
                if self.expanded_extract_into is not None:
                    target_dir = self.expanded_extract_into
                elif self.extract.created_directory is not None:
                    target_dir = os.path.join(os.path.dirname(self.archive_filepath),
                                              self.extract.created_directory)
                else:
                    target_dir = self.archive_filepath[:-len(ext)]
                break
        else:
            # Is run when the loop does not break.
            raise error.AxiiError(
                f"Unknown archive type: '{os.path.splitext(self.archive_filename)}'.\n"
                f"Known are: " + ", ".join(extractors))

        assert None not in [fn, target_dir], [fn, target_dir]

        if refresh or not os.path.isdir(target_dir):
            fn()
            assert os.path.isdir(target_dir), f"{target_dir} does not exist or is not a directory"
            return True
        else:
            return False

    def extract_tar(self):
        import tarfile
        try:
            with tarfile.open(self.archive_filepath) as file:
                file.extractall(self.ctx.path)
        except tarfile.ReadError as e:
            raise error.AxiiError(f"Extracting {self.archive_filepath} failed") from e

    def extract_zip(self):
        import zipfile
        with zipfile.ZipFile(self.archive_filepath) as file:
            file.extractall(self.ctx.path)

    def extract_deb(self):
        try:
            path = self.expanded_extract_into
        except:
            path = None
        if path is None:
            path = os.path.join(self.ctx.path, self.archive_filepath[:-len('.deb')])
        commands.run(f"dpkg -x {self.archive_filepath} {path}", quiet=True, capture_output=True)

    def matches_extension(self, extension: str) -> bool:
        if self.extract.type is not None and extension.endswith(self.extract.type):
            return True

        elif self.archive_filename.endswith(extension):
            return True

        else:
            return False


@dc.dataclass
class Archives(Hook):
    archives: List[Archive]

    def __post_init__(self):
        from armarx_setup.core.util.dataclass import sanitize_type
        if self.archives is None:
            self.archives = []
        else:
            self.archives = [sanitize_type(a, Archive, ctx=self.ctx) for a in self.archives]

    def __call__(self, *args, **kwargs):
        step_results = []
        results = []
        for archive in self.archives:
            step_result = archive()
            if step_result.status == common.StepResultStatus.FAILURE:
                return step_result
            if step_result.status != common.StepResultStatus.SKIPPED:
                step_results.append(step_result)
            results.append(step_result.summary)

        summaries = [res.summary for res in step_results]

        # Eliminate duplicates while preserving order.
        summaries = OrderedDict.fromkeys(summaries)
        if len(summaries) > 0:
            maybe_s = "s" if len(summaries) > 1 else ""
            summary = f"Downloaded and handled {len(summaries)} archive{maybe_s}"
        else:
            summary = "Nothing to do"
        result = "\n".join([f"[i]Archive '{a.archive_filename}':[/] \n{r}"
                            for a, r in zip(self.archives, results)])

        status = common.StepResult.from_sub_steps_results(sub_steps_results=step_results).status

        return common.StepResult(status=status, summary=summary, output=result)


@dc.dataclass
class Update(Hook):
    download: Optional[Download] = None
    git: Optional[Git] = None
    archive: Optional[Archive] = None
    archives: Optional[Archives] = None

    def __post_init__(self):
        from armarx_setup.core.util.dataclass import sanitize_type

        kwargs = { "ctx": self.ctx }
        sub_step_enabled_kwargs = { "list_factory": common.SubStepAdapter.list_factory, **kwargs }

        self.download = sanitize_type(self.download, Download, **sub_step_enabled_kwargs)
        self.git = sanitize_type(self.git, Git, **sub_step_enabled_kwargs)
        self.archive = sanitize_type(self.archive, Archive, **sub_step_enabled_kwargs)
        archives = self.archives
        if archives is not None:
            archives = {"archives": archives}
        self.archives = sanitize_type(archives, Archives, **kwargs)

    def _get_sub_hooks(self) -> List["Hook"]:
        return [self.download, self.git, self.archive, self.archives]
