import os
import networkx as nx
import typing as ty

from typing import Dict, Iterable, Optional, Union

from armarx_setup.core import error
from armarx_setup.core.module import Module, Dependency
from armarx_setup.core.config import axii_config


class ModuleGraph(nx.DiGraph):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_modules(
            cls,
            modules: Union[Iterable[Union[str, Module]], Dict[str, Dict]],
            ws: Optional["Workspace"] = None,
    ):

        self = cls()

        # Add modules that are explicitly listed in the workspace config.
        # We need to add them first without dependencies so their extensions are applied correctly.
        # (If we added a module B as dep. of A before we added B itself, we would not apply the extension to B.)
        if isinstance(modules, dict):
            for name, extension in modules.items():
                self.add_module(name, ws=ws, extension=extension, recursive=False)
        else:
            for module in modules:
                self.add_module(module, ws=ws, recursive=False)

        # Add dependencies (recursively).
        # (Fetch the list of modules beforehand so we do not add new nodes during iteration.)
        modules = [module for _, module in self.nodes(data="module")]
        for module in modules:
            self.add_dependencies(module=module, ws=ws, recursive=True)

        # Add edges by implicitly enabled features.
        for _, module in self.nodes(data="module"):
            module: Module
            for feature_name, required_modules in module.features.items():
                required_modules = [required_modules] if isinstance(required_modules, str)\
                    else required_modules
                # Check whether all modules required by this feature are active.
                if all(self.has_node(required_module) for required_module in required_modules):
                    edges = [(module.name, required_module) for required_module in required_modules]
                    for edge in edges:
                        if not self.has_edge(*edge):
                            # We should respect the optional dependency even
                            # though the feature was not enabled explicitly.
                            # => Consider the feature be enabled.
                            dependency = Dependency(feature_name=feature_name)
                            module.required_modules[edge[1]] = dependency
                            self.add_edge(*edge, dependency=dependency)

                        assert self.has_edge(*edge)

        return self

    def add_module(
            self,
            module: Union[str, Module],
            ws: Optional["Workspace"] = None,
            extension: Optional[Dict] = None,
            requested_features=(),
            recursive=False,
    ):
        console = axii_config.console
        requested_features = list(requested_features)

        if isinstance(module, str):

            if module in self:
                module = self.nodes[module]["module"]
            else:
                if extension is not None:
                    features = extension.pop("features", [])
                    requested_features += features
                try:
                    module = Module.load_info(module, ws=ws, extension=extension)
                except error.ModuleInfoNotFound:
                    module = Module(name=module)
                    module.not_found = True

        # Add module.
        if module.name not in self:
            self.add_node(module.name, module=module)

        # Add required modules if certain features were enabled for this module.  Only relevant if the module was fount.
        if not module.not_found:
            for feature_name in requested_features:
                if feature_name not in module.features:
                    if len(module.features) == 0:
                        console.print(f"Requested feature '{feature_name}' of module '{module.name}', but the module "
                                      f"does not have any feature.")
                    else:
                        available_features = ', '.join(module.features)
                        console.print(f"Requested unknown feature '{feature_name}' of module '{module.name}'.  "
                                      f"Available features are: {available_features}.")
                    continue
                required_name = module.features[feature_name]
                if isinstance(required_name, list):
                    for required_n in required_name:
                        module.required_modules[required_n] = Dependency(feature_name=feature_name)
                else:
                    module.required_modules[required_name] = Dependency(feature_name=feature_name)

        # Add requirements generated by hooks (e.g. git, cmake).
        for phase in module.get_phases():
            if phase is not None:
                for requirement in phase.generate_requirements():
                    module.required_modules[requirement] = Dependency()

        if recursive:
            # Add required dependencies.
            self.add_dependencies(module, ws=ws, recursive=recursive)

    def add_dependencies(self, module: Module, ws: "Workspace", recursive=False):
        for dependency_name, dependency in module.required_modules.items():
            if dependency_name == module.name:
                raise error.ModuleRequiresItself(module.name)

            self.add_module(dependency_name, ws=ws, requested_features=dependency.features,
                            recursive=recursive)

            edge = (module.name, dependency_name)
            if edge not in self.edges:
                self.add_edge(*edge, dependency=dependency)

    def get_module_with_name(self, name: str) -> Module:
        try:
            node = self.nodes[name]
        except KeyError:
            similar = Module.get_similar_module_names(name=name, other_names=self.nodes)
            raise error.NoSuchModule(name, similar_modules=similar)
        else:
            return node["module"]

    def get_dependency_chain(self) -> Iterable[Module]:
        return self._get_dependency_chain(graph=self)

    def get_dependency_chain_to_module(self, module: Module):
        nodes_of_subgraph: ty.Set[str] = nx.descendants(self, module.name)
        # Add the target module (not part of the descendants).
        nodes_of_subgraph.add(module.name)

        subgraph = self.subgraph(nodes_of_subgraph)
        return self._get_dependency_chain(graph=subgraph)

    @classmethod
    def _get_dependency_chain(cls, graph) -> Iterable[Module]:
        try:
            module_names = nx.topological_sort(graph.reverse(copy=True))
        except nx.NetworkXUnfeasible as e:
            raise error.CyclicDependencies(e)
        else:
            chain = [graph.nodes[name]["module"] for name in module_names]
            return chain

    @property
    def layouts(self):
        if self._layouts is None:
            layouts = dict()

            def mutlipartite(graph):
                subsets: Dict[str, int] = dict()
                for n, attr in graph.nodes.items():
                    key = os.path.split(attr["module"].name)[0]
                    if key not in subsets:
                        subsets[key] = len(subsets)
                    attr["subset"] = subsets[key]
                return nx.layout.multipartite_layout(graph, subset_key="subset", align="horizontal")

            layouts["multipartite"] = mutlipartite

            layouts["planar"] = lambda graph: nx.planar_layout(graph)
            layouts["shell"] = lambda graph: nx.shell_layout(graph)
            layouts["spectral"] = lambda graph: nx.spectral_layout(graph)
            layouts["spring"] = lambda graph: nx.spring_layout(graph)

            # Requires apt package "graphviz" and python package "pygraphviz"
            # sudo apt install graphviz
            # For Python >3.7: pip install pygraphviz
            # For Python 3.6: pip install pygraphviz==1.6
            from networkx.drawing.nx_agraph import graphviz_layout
            layouts["graphviz"] = lambda graph: graphviz_layout(graph, prog="dot")

            self._layouts = layouts
        return self._layouts

    def visualize(self, **kwargs):
        from .plotting import plot
        plot(self, **kwargs)

    def filter_hidden_modules(self, modules: Iterable[Union[str, Module]]):
        return list(filter(lambda n: not self.is_hidden_module(n), modules))

    def is_hidden_module(self, module: Union[str, Module]):
        if isinstance(module, str):
            module = self.get_module_with_name(module)
        return any([
            module.name.startswith("tools/"),
            module.name.startswith("apt/"),
            module.is_system_module(),
        ])

    def find_module_explaining_path(self, path: str, ws_path: str):
        def remove_prefix(string: str, prefix: str) -> str:
            assert string.startswith(prefix), (string, prefix)
            return string[len(prefix):]

        path = os.path.abspath(os.path.expandvars(path))

        if not path.startswith(ws_path):
            raise error.AxiiError(
                f"The path '{path}' is not in the active workspace '{ws_path}'.")

        path_in_ws = remove_prefix(path, ws_path)

        if path_in_ws:
            matching_modules = []
            for name, module in self.nodes.data("module"):
                module_path_in_ws = remove_prefix(module.path, ws_path)
                if path_in_ws.startswith(module_path_in_ws):
                    matching_modules.append(module)

            if matching_modules:
                # Take deepest module.
                module = max(matching_modules, key=lambda m: len(m.path))
                return module

        raise error.AxiiError(f"Found no module explaining the path '{path}'.")
