import os
import re
import typing as ty

from armarx_setup import console


def expand_vars(
        text: str,
        default_value: ty.Optional[str] = None,
        consider_escapes: bool = False
) -> str:
    """
    Expands environment variables in given text.

    Taken from StackOverflow: https://stackoverflow.com/a/30777398

    Parameters
    ----------
    text: str
        Text with environment variables to be expanded.
    default_value: ty.Optional[str] = None
        Default value that should be used if the environment variable is not set.  If set to None
        (default),  The variable will be left unchanged and unresolved.  To emulate shell behavior,
        set this to the empty string.
    consider_escapes: bool = False
        Set to true if backslash-escaped environment variables should be considered.  To emulate
        shell behavior, set this to True.

    Returns
    -------
    str
        Input text with expanded environment variables as specified.
    """

    def replace_var(m):
        return os.environ.get(m.group(2) or m.group(1),
                              m.group(0) if default_value is None else default_value)

    re_var = (r'(?<!\\)' if consider_escapes else '') + r'\$(\w+|\{([^}]*)\})'
    return re.sub(re_var, replace_var, text)


def shell_expand_vars(text: str) -> str:
    return expand_vars(text, default_value="", consider_escapes=True)


class EnvironContext:

    def __init__(
            self,
            env: ty.Optional[ty.Dict[str, str]],
            expand_environment_variables: bool = False,
            verbose=-1
    ):
        self.env = env
        self.old_env = None
        self.expand_environment_variables = expand_environment_variables
        self.verbose = verbose

    def __enter__(self) -> ty.Optional["EnvironContext"]:
        if self.env is None or len(self.env) == 0:
            return None
        self.old_env = {}
        for key, value in self.env.items():
            if key in os.environ:
                self.old_env[key] = os.environ[key]

            if self.expand_environment_variables:
                os.environ[key] = shell_expand_vars(value)
            else:
                os.environ[key] = value

        # Print only after environment has been set up for expansion.
        if self.verbose >= 0:
            console.print(f"Set env vars: {self._print_env_dict(self.env)} ...")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore environment
        if self.env is None or len(self.env) == 0:
            return
        if self.verbose >= 0:
            deleted = sorted(set(self.env).difference(set(self.old_env)))
            console.print(f"Restore env vars: {self._print_env_dict(self.old_env)} "
                          f"(delete {', '.join(deleted)}) ...")
        for key, _value in self.env.items():
            if key in self.old_env:
                os.environ[key] = self.old_env[key]
            else:
                del os.environ[key]

    def print_env_dict(self):
        return self._print_env_dict(self.env)

    @staticmethod
    def _print_env_dict(env: ty.Dict[str, str]) -> str:
        items = []
        for key, raw_value in env.items():
            exp_value = os.environ.get(key, '?')
            expansion = ""
            if raw_value != exp_value:
                expansion = f" (='{exp_value}')"
            items.append(f"{key}='{raw_value}'{expansion}")
        return " ".join(items)
