"""Lints an Axii module definition for compliance with conventions."""

import dataclasses as dc
import itertools
import json
import hashlib
import typing as ty

import json_source_map

import armarx_setup
from armarx_setup.core.module import Module


def md5(string):
    return hashlib.md5(string.encode("utf-8")).hexdigest()


@dc.dataclass
class Issue:
    description: str = ""
    fingerprint: str = ""
    severity: str = ""
    path: str = ""
    line: int = 0


INFO = "info"
MINOR = "minor"
MAJOR = "major"
CRITICAL = "critical"
BLOCKER = "blocker"


class ModuleLinter:

    def __init__(self, module_name: str):
        self.module_name = module_name
        self.module_path = Module.get_info_path(module_name)
        with open(self.module_path, "r", encoding="utf-8") as file:
            json_str = file.read()
            try:
                self.module_json = json.loads(json_str)
                self.module_source_map = json_source_map.calculate(json_str)
            except json.decoder.JSONDecodeError:
                self.module_json = None
                self.module_source_map = None
        self.steps = ["system", "update", "prepare", "build", "install", "conclude"]
        self.sections_in_order = ["general"] + self.steps + ["features", "required_modules"]
        self.issues = []

    def lint(self):
        if self.module_json is None:
            self._report_issue(
                description="Could not parse JSON file due to syntax errors",
                severity=BLOCKER,
            )
        else:
            # Overall.
            self.lint_sections()
            self.lint_new_lines()
            self.lint_sections_order()

            # Section-specific,
            self.lint_general_section()
            self.lint_install_section()
        return self.issues

    def lint_sections(self):
        for section in self.module_json.keys():
            if section not in self.sections_in_order and not section.startswith("#"):
                self._report_issue(
                    description=f"Unknown section '{section}'",
                    severity=CRITICAL,
                    line=self._key_start(section)
                )

    def lint_new_lines(self):
        expected_line_diff = 2
        order_severity = MAJOR

        if len(self.module_json.keys()) == 0:
            return

        sections = list(self.module_json.keys())
        previous_section = sections[0]
        for current_section in sections[1:]:
            previous_section_end = self._value_end(previous_section)
            current_section_start = self._key_start(current_section)

            if previous_section_end + expected_line_diff != current_section_start:
                self._report_issue(
                    description=f"Expected exactly one empty new line between '{previous_section}' "
                                f"and '{current_section}' sections",
                    severity=order_severity,
                    line=self._value_end(previous_section) + 1
                )

            previous_section = current_section

    def lint_sections_order(self):
        positions: ty.Dict[str, ty.Optional[str]] = {}

        for section in self.sections_in_order:
            if f"/{section}" in self.module_source_map:
                positions[section] = self._key_start(section)
            else:
                positions[section] = None

        offending_sections = {}

        for sec_1, sec_2 in itertools.combinations(self.sections_in_order, 2):
            sec_1: str
            sec_2: str
            if positions[sec_1] is None or positions[sec_2] is None:
                continue

            if positions[sec_2] < positions[sec_1]:
                offending_sections[sec_2] = dict(
                    description=f"Section '{sec_2}' should come after section '{sec_1}'.",
                    severity=MAJOR,
                    line=positions[sec_1]
                )

        for _, value in offending_sections.items():
            self._report_issue(**value)

    def lint_general_section(self):
        if "general" not in self.module_json.keys():
            self._report_issue(
                description="No 'general' section declared",
                severity=MAJOR,
                line=1
            )
        else:
            self.lint_general_section_contents()

    def lint_general_section_contents(self):
        general_section_line = self._key_start("general")

        if "url" not in self.module_json["general"].keys():
            self._report_issue(
                description="No 'url' subsection provided in 'general' section",
                severity=MAJOR,
                line=general_section_line
            )

        if "authors" not in self.module_json["general"].keys():
            self._report_issue(
                description="No 'authors' subsection in 'general' section declared",
                severity=MAJOR,
                line=general_section_line
            )
        else:
            from email.utils import parseaddr

            authors = self.module_json["general"]["authors"]
            if isinstance(authors, str):
                authors = [authors]

            for author in authors:
                display_name, mail_address = parseaddr(author)
                if mail_address == "" or display_name == "":
                    self._report_issue(
                        description=f"Author notice '{author}' malformed",
                        severity=MAJOR,
                        line=self._value_start("general/authors")
                    )

    def lint_install_section(self):
        if "install" not in self.module_json.keys():
            return

    def _report_issue(self, **kwargs):
        if "path" not in kwargs:
            kwargs["path"] = self.module_path.replace(f"{armarx_setup.PACKAGE_ROOT}/", "")
        if "fingerprint" not in kwargs:
            kwargs["fingerprint"] = md5(kwargs["path"] + kwargs["description"])
        self.issues.append(Issue(**kwargs))

    def _key_start(self, key):
        return self.module_source_map[f"/{key}"].key_start.line + 1

    def _key_end(self, key):
        return self.module_source_map[f"/{key}"].key_end.line + 1

    def _value_start(self, key):
        return self.module_source_map[f"/{key}"].value_start.line + 1

    def _value_end(self, key):
        return self.module_source_map[f"/{key}"].value_end.line + 1
