import os
import subprocess
import typing as ty
from typing import Dict, List

from armarx_setup.core import error

apt_package_cache: Dict[str, bool] = {}
flatpak_package_cache: ty.Dict[ty.Tuple[str, str], bool] = {}


def set_process_priorities(nice: int = 0, ionice_class="none"):
    try:
        import psutil

        ionice_map = {
            "real_time": psutil.IOPRIO_CLASS_RT,
            "best_effort": psutil.IOPRIO_CLASS_BE,
            "idle": psutil.IOPRIO_CLASS_IDLE,
            "none": psutil.IOPRIO_CLASS_NONE
        }

        assert ionice_class in ionice_map, f"ionice_class must be any of {', '.join(ionice_map.keys())}"

        axii_process = psutil.Process(os.getpid())
        axii_process.nice(nice)
        axii_process.ionice(psutil.IOPRIO_CLASS_IDLE)
    except:
        pass  # Ignore error silently for now, Axii will just not behave so nicely in this case.


def is_apt_package_installed(package_name: str) -> bool:
    if not apt_package_cache:
        apply_dpkg_list()

    if package_name in apt_package_cache:
        if package_name == "screen":
            print("screen:", apt_package_cache[package_name])
        return apt_package_cache[package_name]
    else:
        # This can indicate that a package is installed although it has just been removed via apt-get remove.
        # dpkg lists the packages as "rc" (instead of "ii" for installed), but "-V ..." is still successful.
        is_installed = os.system(f'dpkg -V "{package_name}" > /dev/null 2>&1') == 0
        apt_package_cache[package_name] = is_installed
        return is_installed


def is_flatpak_package_installed(origin: str, package_name: str) -> bool:
    from armarx_setup.core.util import commands

    package_id = (origin, package_name)

    if package_id in flatpak_package_cache:
        return flatpak_package_cache[package_id]

    is_installed = False
    matches_origin = False

    def check_origin(line: str):
        nonlocal matches_origin
        line = commands.strip_ansi(line).strip()
        try:
            key, value = line.split(": ", maxsplit=1)
        except ValueError:
            error.handle_exception()
        else:
            if key == "Origin":
                matches_origin = value == origin

    try:
        commands.execute(["flatpak", "info", package_name],
                         stdout_callback=check_origin)
    except error.CommandFailed:
        error.handle_exception()
        is_installed = False
    else:
        is_installed = True

    state = is_installed and matches_origin
    flatpak_package_cache[package_id] = state
    return state


def apply_dpkg_list():
    global apt_package_cache

    process = subprocess.run("dpkg --list", shell=True, stdout=subprocess.PIPE)
    out = process.stdout.decode()
    lines = out.split("\n")

    def get_names_with_prefix(lines: List[str], prefix: str) -> List[str]:
        filtered = [line for line in lines if line.startswith(prefix)]
        table = [list(filter(bool, line.split(" "))) for line in filtered]
        names = [row[1] for row in table]
        return names

    installed = get_names_with_prefix(lines=lines, prefix="ii")
    removed = get_names_with_prefix(lines=lines, prefix="rc")

    for name in installed:
        apt_package_cache[name] = True
    for name in removed:
        apt_package_cache[name] = False


def flush_apt_cache():
    global apt_package_cache
    apt_package_cache = {}


def flush_flatpak_cache():
    global flatpak_package_cache
    flatpak_package_cache = {}


def install_apt_packages(packages: List[str], accept=False):
    from armarx_setup.core.util import cli, commands

    packages = sorted(set(packages))
    s_packages = "package" + ('s' if len(packages) > 1 else '')

    user_id = os.getuid()
    is_root = user_id == 0
    sudo = "" if is_root else "sudo"

    wss_list = ' '.join(packages)
    command = f"{sudo} apt-get install --no-install-recommends -y {wss_list}".lstrip()

    from armarx_setup.cli import common
    if accept or common.input_confirm(
            f"Run '{command}' to install {len(packages)} {s_packages}?",
            default_option=False,
    ):
        flush_apt_cache()
        commands.run(command)


def install_flatpak_packages(origin: str, packages: List[str], accept=False):
    from armarx_setup.core.config import axii_config
    execute = axii_config.executor

    packages = sorted(set(packages))
    s_packages = "package" + ("s" if len(packages) > 1 else "")

    wss_list = " ".join(packages)
    command = f"flatpak install {origin} {wss_list} -y".lstrip()

    from armarx_setup.cli import common
    if accept or common.input_confirm(
            f"Run '{command}' to install {len(packages)} {s_packages}?",
            default_option=False,
    ):
        flush_flatpak_cache()
        execute(command.split())
