import os
import time
import datetime
import dataclasses as dc
import typing as ty

from collections import OrderedDict

from armarx_setup import console

DEFAULT_FILENAME = "armarx-workspace.rc"


@dc.dataclass
class Variable:
    name: str
    value: str
    function: str = "_axii_env_set"

    def make_line(self) -> str:
        return f'{self.function} {self.name} "{self.value}"'


@dc.dataclass
class Source:
    path: str

    def make_line(self) -> str:
        lines = [
            f'if [ -f "{self.path}" ]; then',
            f'    source "{self.path}"',
            f'fi'
        ]
        return "\n".join(lines)


@dc.dataclass
class Entry:
    variables: ty.List[Variable] = dc.field(default_factory=list)
    sources: ty.List[Source] = dc.field(default_factory=list)

    def make_lines(self) -> ty.List[str]:
        lines = []
        lines += [var.make_line() for var in self.variables]
        lines += [source.make_line() for source in self.sources]
        return lines


class EnvFile:

    def __init__(
            self,
            ws_path: str,
            filename: str = DEFAULT_FILENAME,
    ):

        self.env: OrderedDict[str, Entry] = OrderedDict()

        self.ws_path = ws_path
        self.filename = filename

    def add(self, name: str, value: str, by: str, write=False, quiet=False):
        value = str(value)

        # If ":$<variable name>" or "$<variable name>:" is included in the value, we have a prepend/append.
        if f":${name}" in value or f"${name}:" in value:
            # Split the values which are prepended or appended to the variable.
            prepends, appends = value.split(f"${name}")
            parts = filter(None, prepends.split(":"))
            for part in parts:
                if not quiet:
                    console.print(f"Prepend '{part}' to {name}.")
                self.env.setdefault(by, Entry()).variables.append(Variable(name=name, value=part,
                                                                           function="_axii_env_prepend_to"))
            parts = filter(None, appends.split(":"))
            for part in parts:
                if not quiet:
                    console.print(f"Append '{part}' to {name}.")
                self.env.setdefault(by, Entry()).variables.append(Variable(name=name, value=part,
                                                                           function="_axii_env_append_to"))
        else:
            if not quiet:
                console.print(f"Set {name} to '{value}'.")
            self.env.setdefault(by, Entry()).variables.append(Variable(name=name, value=value))

        os.environ[name] = os.path.expandvars(value)

        if write:
            self.write()

    def remove(self, name: str, by: str, write=False, quiet=False):
        if not quiet:
            console.print(f"Unset ${name}.")
        # Remove the variable from the environment file by removing its entry from the env dictionary.
        self.env[by].variables = [var for var in self.env[by].variables if var.name != name]

        if write:
            self.write()

    def add_source(self, path: str, by: str, write=False):
        path = str(path)
        self.env.setdefault(by, Entry()).sources.append(Source(path=path))

        console.print(f"Source '{path}'.")

        if write:
            self.write()

    def get_path(self) -> str:
        return os.path.join(self.ws_path, self.filename)

    def get_header(self) -> ty.List[str]:
        now = datetime.datetime.fromtimestamp(int(time.time()))
        return [
            f"# This environment file was automatically generated by Axii on {now}.",
            "",
            "",
            "_axii_env_reset",
            "",
            ""
        ]

    def make_by(self, by_name: str) -> str:
        return f"# Generated by '{by_name}'"

    def get_lines(self) -> ty.List[str]:
        lines = []

        lines += self.get_header()

        for by, entry in self.env.items():
            lines.append(self.make_by(by))
            lines += entry.make_lines()
            lines.append("")

        return lines

    def write(self):
        lines = self.get_lines()
        with open(self.get_path(), "w") as file:
            file.write("\n".join(lines) + "\n")
