import enum
from typing import Dict, List, Optional, Union

from armarx_setup import console
from armarx_setup.core import error
from armarx_setup.core.util import prefix_color_map
from armarx_setup.core.module import Module
from .workspace import Workspace


class SystemStep:
    class Type(enum.IntEnum):
        Run = 1
        Packages = 2
        Env = 3
        Source = 4

        @classmethod
        def of_module(cls, module: Module) -> "SystemStep.Type":
            if module.system.run:
                return cls.Run
            elif module.system.packages:
                return cls.Packages
            elif module.system.env:
                return cls.Env
            elif module.system.source:
                return cls.Source
            else:
                raise ValueError(f"Unhandled module {module.name} by enum {cls.__name__}.")

    def __init__(
            self,
            type: Type,
            modules: List[Module] = None,
            ws: Optional[Workspace] = None,
    ):
        self.type = type
        self.modules = modules or []
        self.ws = ws

    def __repr__(self):
        return "<{c} type={t} #modules={nm}>".format(
            c=self.__class__.__name__,
            t=self.type,
            nm=len(self.modules)
        )

    def run(self):
        assert self.type == self.Type.Run
        for module in self.modules:
            assert module.system is not None
            module.system.run()

    def install_packages(self, accept=False):
        assert self.type == self.Type.Packages

        apt_packages = self.gather_apt_packages()

        if apt_packages["found"]:
            found: Dict[str, List[Module]] = apt_packages["found"]
            Module.say_num_packages_are_installed(len(found), package_manager="apt")
            console.print()
            console.print(self.to_table_packages_and_dependent_modules(found, package_manager="Apt"))

        if apt_packages["missing"]:
            from armarx_setup.core.util import system

            missing: Dict[str, List[Module]] = apt_packages["missing"]

            console.print()
            console.print(
                f"The following {len(missing)} apt packages are required but not installed yet:")
            console.print(self.to_table_packages_and_dependent_modules(missing, package_manager="Apt"))

            system.install_apt_packages(list(missing), accept=accept)

        flatpak_packages_per_origin = self.gather_flatpak_packages()

        for origin, flatpak_packages in flatpak_packages_per_origin.items():
            if flatpak_packages["found"]:
                found: Dict[str, List[Module]] = flatpak_packages["found"]
                Module.say_num_packages_are_installed(len(found), package_manager=f"flatpak ({origin})")
                console.print()
                console.print(self.to_table_packages_and_dependent_modules(found, package_manager="Flatpak"))

            if flatpak_packages["missing"]:
                from armarx_setup.core.util import system

                missing: Dict[str, List[Module]] = flatpak_packages["missing"]

                console.print()
                console.print(
                    f"The following {len(missing)} flatpak packages from {origin} are required but not installed yet:")
                console.print(self.to_table_packages_and_dependent_modules(missing, package_manager="Flatpak"))

                system.install_flatpak_packages(origin=origin, packages=list(missing), accept=accept)

    def gather_apt_packages(
            self,
            max_workers=8,
            verbose=1,
    ) -> Dict[str, Union[Dict[str, List[Module]], List[Module]]]:
        from concurrent import futures
        from rich.live import Live

        def run_jobs(live: Optional[Live] = None):
            jobs = []
            with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                def job(module: Module):

                    if live is not None:
                        module_text = prefix_color_map.module_to_colored_text(module.name)
                        live.update(f"Checking apt packages of module {module_text} ...")

                    checked: Dict[str, List[str]] = module.system.check_apt_packages()
                    return checked, module

                for module in self.modules:
                    assert module.system
                    assert module.system.packages
                    jobs.append(executor.submit(job, module))

            return jobs

        if verbose >= 1:
            with Live("", refresh_per_second=30) as live:
                jobs = run_jobs(live=live)
        else:
            jobs = run_jobs()

        gathered = {
            "found": {},
            "missing": {}
        }
        for job in jobs:
            checked, module = job.result()
            if checked is None:
                continue
            for key in checked:
                packages_dict = gathered[key]
                for package in checked[key]:
                    if package not in packages_dict:
                        packages_dict[package] = []
                    packages_dict[package].append(module)

        return gathered

    def gather_flatpak_packages(
            self,
            max_workers=8,
            verbose=1,
    ) -> Dict[str, Dict[str, Union[Dict[str, List[Module]], List[Module]]]]:
        from concurrent import futures
        from rich.live import Live

        def run_jobs(live: Optional[Live] = None):
            jobs = []
            with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                def job(module: Module):

                    if live is not None:
                        module_text = prefix_color_map.module_to_colored_text(module.name)
                        live.update(f"Checking flatpak packages of module {module_text} ...")

                    checked: Optional[Dict[str, Dict[str, List[str]]]] = module.system.check_flatpak_packages()
                    return checked, module

                for module in self.modules:
                    assert module.system
                    assert module.system.packages
                    jobs.append(executor.submit(job, module))

            return jobs

        if verbose >= 1:
            with Live("", refresh_per_second=30) as live:
                jobs = run_jobs(live=live)
        else:
            jobs = run_jobs()

        gathered_per_origin = {}

        for job in jobs:
            checked, module = job.result()
            if checked is None:
                continue
            for origin, packages in checked.items():
                if origin not in gathered_per_origin:
                    gathered_per_origin[origin] = {
                        "found": {},
                        "missing": {}
                    }
                for key in ["found", "missing"]:
                    packages_dict = gathered_per_origin[origin][key]
                    for package in packages[key]:
                        if package not in packages_dict:
                            packages_dict[package] = []
                        packages_dict[package].append(module)

        return gathered_per_origin

    def to_table_packages_and_dependent_modules(
            self,
            packages_and_dependent_modules: Dict[str, List[Module]],
            package_manager: str
    ):
        from rich.table import Table
        table = Table("#", f"{package_manager} Package Name", "Module(s)", "Directly Dependent Modules")
        table.columns[0].justify = "right"

        for i, (package, modules) in enumerate(sorted(packages_and_dependent_modules.items())):
            dependent_modules = set()
            for m in modules:
                for (source, target) in self.ws.module_graph.in_edges(m.name):
                    dependent_modules.add(source)
            table.add_row(
                str(i + 1),
                package,
                ", ".join(
                    prefix_color_map.modules_to_colored_text(sorted([m.name for m in modules]))),
                ", ".join(prefix_color_map.modules_to_colored_text(sorted(dependent_modules)))
            )
        return table


class System:

    def __init__(self, ws: Workspace):
        self.ws = ws

    def plan_steps(self, modules: List[Module]) -> List[SystemStep]:

        steps: List[SystemStep] = [
            # Start with a run step, if required.
            SystemStep(type=SystemStep.Type.Run, ws=self.ws),
        ]

        def find_dependency(module_name: str):
            # This must always succeed as dependencies are processed before their dependents.
            # next(filter()) implements a find_if()
            filtered = filter(lambda i_step: module_name in [m.name for m in i_step[1].modules],
                              enumerate(steps))
            try:
                return next(filtered)
            except StopIteration:
                console.print(f"module that was depended on: {module_name}")
                raise

        def check_dependencies(module_name_to_insert: str, target_step_index: int):
            for dep in self.ws.module_graph.adj[module_name_to_insert]:
                if self.ws.module_graph.nodes[dep]["module"].system is None:
                    module = f"'{module_name_to_insert}'"
                    suggested_new_module = f"'{module_name_to_insert}-system-requirements"
                    lines = [
                        f"The module {module} specifies a system step, "
                        f"but requires '{dep}', which does not specify a system step.",
                        f"System modules may only require modules with a system step. "
                        f"(Otherwise, respecting requirements is not possible when setting up the system up-front).",
                        f"To fix this issue:",
                        f"If {module} specifies a system/packages/apt step:",
                        f"  (1.) For each apt package 'package', add a required module 'apt/package' to {module}.",
                        f"       If there is no corresponding module description (.json file in the module database),",
                        f"       the required 'apt/...' module will be emulated dynamically.",
                        f"  (2.) Remove the system/packages/apt step.",
                        f"If there are other system steps in {module}:",
                        f"  (1.) Create a new module, e.g. {suggested_new_module}.",
                        f"  (2.) Move the system step from {module} to {suggested_new_module}.",
                        f"  (3.) Add {suggested_new_module} to the required modules of {module}.",
                    ]
                    raise error.AxiiError("\n".join(lines))

                # console.print(f"> {dep}")
                dep_i, dep_step = find_dependency(dep)
                if dep_i > target_step_index:
                    return False
            return True

        def add_to_first_fitting_step(module: Module):
            step_type = SystemStep.Type.of_module(module)
            for step_i, step in filter(lambda i_s: i_s[1].type == step_type, enumerate(steps)):
                if check_dependencies(module.name, step_i):
                    step.modules.append(module)
                    return True
            return False

        for i, module in enumerate(modules):
            # print(f"[{i}] {module.name}")
            # console.print(f"[b]{module.name}[/]")
            if not add_to_first_fitting_step(module):
                steps.append(SystemStep(type=SystemStep.Type.of_module(module), modules=[module],
                                        ws=self.ws))

        steps += [
            SystemStep(type=SystemStep.Type.Env,
                       modules=[m for m in modules if m.system.env is not None], ws=self.ws)
        ]

        return steps
