"""
Tests `axii workspace` functionalities.
"""

import os
import pytest

from armarx_setup.core import error
from armarx_setup.core.config import axii_config
from armarx_setup.core.util.commands import run
from armarx_setup.core.workspace import Workspace

from tests.e2e import cmd_args, test_workspace


def test_workspace_create_and_purge():
    """
    Tests creating a workspace and purging it afterwards.
    """

    workspace_name = "test_workspace"
    workspace_path = os.path.join("/tmp", workspace_name)

    assert not os.path.isdir(workspace_path), \
        "Workspace folder should not exist to begin with."

    assert workspace_name not in axii_config.get_known_workspaces(), \
        "Workspace should have been removed from known_workspaces.json."

    stdout = run(f"axii workspace create {workspace_path} {workspace_name}", **cmd_args)
    axii_config.load_known_workspaces()

    assert "Successfully created the workspace." in stdout
    workspace = Workspace.load_workspace_by_path(workspace_path)
    assert os.path.isdir(workspace.path), \
        "Workspace folder should have been created."
    assert os.path.isfile(workspace.get_config_path()), \
        "The file armarx-workspace.json should have been created."
    assert os.path.isfile(os.path.join(workspace.path, "armarx-workspace.rc")), \
        "The file armarx-workspace.rc should have been created."
    assert workspace_name in axii_config.get_known_workspaces(), \
        "Workspace should have been added to known_workspaces.json."

    # Simulate activating workspace.
    os.environ["ARMARX_WORKSPACE"] = workspace_path

    stdout = run(f"axii workspace purge {workspace_name} --yes", **cmd_args)
    axii_config.load_known_workspaces()

    assert "Purged workspace 'test_workspace'." in stdout
    assert "Cleaned up metadata." in stdout
    assert workspace_name not in axii_config.get_known_workspaces(), \
        "Workspace should have been removed from known_workspaces.json."
    assert not os.path.isdir(workspace_path), \
        "Workspace folder should have been deleted."

    # Simulate deactivating workspace.
    del os.environ["ARMARX_WORKSPACE"]


def test_workspace_create_ccache():
    """
    Tests creating a workspace with and without Ccache.
    """

    workspace_name = "test_workspace2"
    workspace_path = os.path.join("/tmp", workspace_name)

    def ccache_in_workspace():
        workspace_file_path = os.path.join(workspace_path, "armarx-workspace.json")
        with open(workspace_file_path, "r", encoding="utf-8") as workspace_file:
            return "tools/ccache/default" in workspace_file.read()

    stdout = run(f"axii workspace create {workspace_path} {workspace_name}", **cmd_args)

    assert not ccache_in_workspace()

    stdout = run(f"axii workspace purge {workspace_name} --yes", **cmd_args)
    stdout = run(f"axii workspace create {workspace_path} {workspace_name} --ccache", **cmd_args)

    assert ccache_in_workspace()

    stdout = run(f"axii workspace purge {workspace_name} --yes", **cmd_args)
    stdout = run(f"axii workspace create {workspace_path} {workspace_name} --no-ccache", **cmd_args)

    assert not ccache_in_workspace()

    stdout = run(f"axii workspace purge {workspace_name} --yes", **cmd_args)


def test_workspace_create_existing():
    """
    Tests that creating a workspace that already exists will result in an error.
    """


def test_workspace_purge_non_existent():
    """
    Tests that purging a non-existent workspace will result in an error.
    """

    workspace_name = "test_workspace"

    assert workspace_name not in axii_config.get_known_workspaces(), \
        f"Workspace '{workspace_name}' should not be in known_workspaces.json."

    try:
        run(f"axii workspace purge {workspace_name} --yes", **cmd_args)
    except error.CommandFailed as e:
        stdout = e.stdout
        return_code = e.return_code
    else:
        return_code = 0
        stdout = ""

    assert return_code == 1, "Return code should be 1."
    assert f"No such workspace '{workspace_name}'." in stdout


def test_workspace_remove_module(test_workspace):
    """
    Tests removing a module from a workspace.
    """

    workspace_config = test_workspace.load_config()
    assert "armarx" in workspace_config.modules

    stdout = run("axii workspace remove armarx", **cmd_args)

    assert "Removed module 'armarx'." == stdout
    workspace_config = test_workspace.load_config()
    assert "armarx" not in workspace_config.modules


def test_workspace_remove_module_twice(test_workspace):
    """
    Tests that removing a module which is not in the workspace will result in an error.
    """

    run("axii workspace remove armarx", **cmd_args)  # Remove 'armarx' from default workspace.

    old_workspace_config = test_workspace.load_config()

    # Remove it again (it is not in the workspace after removing it once).
    stdout = run("axii workspace remove armarx", **cmd_args)

    assert "Module 'armarx' is not in the current workspace." == stdout
    new_workspace_config = test_workspace.load_config()
    assert old_workspace_config.modules == new_workspace_config.modules, \
        "Workspace config before and after removing should be the same."


def test_remove_non_existent_module_from_workspace():
    """
    Tests that removing an non-existent module will result in an error.
    """

    # TODO: Currently same behaviour as removing an existing module. Should be a different error
    # tough.


def test_prepare_python_preinstall(test_workspace: Workspace):
    import shutil

    print(test_workspace)
    assert test_workspace is not None

    stdout = run("axii workspace remove armarx", **cmd_args)
    stdout = run("axii workspace remove h2t/code_style", **cmd_args)

    test_module_name = "test/prepare/python/preinstall"
    stdout = run(f"axii workspace add {test_module_name}", **cmd_args)

    from armarx_setup.core.module import Module
    module_def_path = Module.get_info_path(test_module_name)
    assert os.path.isfile(module_def_path), module_def_path

    package_source_dir = os.path.join(
        os.path.splitext(module_def_path)[0], "preinstall_test")
    assert os.path.isdir(package_source_dir), package_source_dir

    # Unzip it to the test workspace.
    module_path = os.path.join(test_workspace.path, test_module_name)
    package_dest_dir = os.path.join(module_path, "preinstall_test")

    assert not os.path.exists(module_path), module_path
    os.makedirs(module_path, exist_ok=False)
    assert os.path.isdir(module_path), module_path

    shutil.copytree(package_source_dir, package_dest_dir)
    assert os.path.isdir(package_dest_dir), package_dest_dir
    assert os.path.isfile(os.path.join(package_dest_dir, "pyproject.toml"))

    stdout = run(f"axii workspace upgrade", **cmd_args)
    print(stdout)

    venv_dir = os.path.join(package_dest_dir, ".venv")
    assert os.path.isdir(venv_dir), venv_dir

    python_bin = os.path.join(venv_dir, "bin", "python")
    assert os.path.isfile(python_bin), python_bin

    packages = [
        # From preinstall in Axii module definition.
        "packaging",
        # From pyproject.toml.
        "numpy"
    ]
    for package in packages:
        python_command = f"import {package}; print('OK')"
        try:
            stdout = run(f'{python_bin} -c "{python_command}"', capture_output=True)
        except error.CommandFailed as e:
            print("stdout:", e.stdout)
            print("stderr:", e.stderr)
            raise
        else:
            assert "OK" in stdout, stdout
