"""Unit tests for the armarx_setup.core.config module."""

import pytest

from armarx_setup.core import error
# Basic implementations of Config and ConfigVariable.
from armarx_setup.core.config import AxiiConfig, Config, ConfigVariable
# Overrides of ConfigVariable and Config with Axii built-in variables.
from armarx_setup.core.config import ConfigWithBuiltIns, BuiltInConfigVariables, \
    BuiltInConfigVariable
# Overrides of Config for git_remote_preferences.
from armarx_setup.core.config import GitRemotePreferencesSubconfig
# OVerrides of ConfigVariable for build_job_number.
from armarx_setup.core.config import BuildJobNumberConfigVariable


def test_config():
    """Tests basic Config usage."""

    cfg_dict = {
        "some_var": True,
        "some_dict": {
            "option_a": "on",
            "option_b": "off",
            "option_c": "on"
        },
        "some_int": 42
    }

    cfg = Config.from_dict(**cfg_dict)

    assert cfg.get("some_var") is True
    assert cfg.get("some_int") == 42

    with pytest.raises(error.NoSuchConfigVariable):
        cfg.get("non_existent_variable")

    cfg_var: Config = cfg.get("some_dict")

    assert type(cfg_var) is Config
    assert cfg_var.get("option_a") == "on"


def test_interpret_build_job_number_valid():
    """Valid values for build_job_number."""

    cfg_var = BuildJobNumberConfigVariable()

    assert cfg_var.interpret_value("1") == 1
    assert cfg_var.interpret_value("4") == 4
    assert cfg_var.interpret_value("all") >= 1
    assert cfg_var.interpret_value("all_but_one") >= 1
    assert cfg_var.interpret_value("half") >= 1
    assert cfg_var.interpret_value("all_coop") >= 1
    assert cfg_var.interpret_value("all_but_one_coop") >= 1
    assert cfg_var.interpret_value("half_coop") >= 1


def test_interpret_build_job_number_invalid():
    """Invalid values for build_job_number expecting exceptions."""

    cfg_var = BuildJobNumberConfigVariable()

    with pytest.raises(error.InvalidConfigurationValue):
        cfg_var.interpret_value("0")
    with pytest.raises(error.InvalidConfigurationValue):
        cfg_var.interpret_value("-1")
    with pytest.raises(error.InvalidConfigurationValue):
        cfg_var.interpret_value("everything")
    with pytest.raises(error.InvalidConfigurationValue):
        cfg_var.interpret_value("everything_coop")


def test_built_in_config_variables():
    """Test built-in config variables."""

    built_in_cfgs = BuiltInConfigVariables.get_all_names()

    assert "dev_mode" in built_in_cfgs
    assert "default_workspace" in built_in_cfgs

    cfg: BuiltInConfigVariable = BuiltInConfigVariables.get_by_name("dev_mode")

    assert cfg.default_value is False

    cfg2: BuiltInConfigVariable = BuiltInConfigVariables.dev_mode

    assert not cfg is cfg2  # Should be a copy, not the canonical built-in.


def test_interpret_git_remote_preferences():
    """Test the git_remote_preferences built-in config variable."""

    cfg_dict = {
        "git_remote_preferences": {
            "git.h2t.iar.kit.edu": "ssh",
            "gitlab.com": "ssh",
            "github.com": "https"
        }
    }

    cfg = ConfigWithBuiltIns.from_dict(**cfg_dict)

    git_remote_cfg_var: GitRemotePreferencesSubconfig = cfg.get("git_remote_preferences")

    assert type(git_remote_cfg_var) is GitRemotePreferencesSubconfig

    assert git_remote_cfg_var.preference_for_host("git.h2t.iar.kit.edu") == "ssh"
    assert git_remote_cfg_var.preference_for_host("gitlab.com") == "ssh"
    assert git_remote_cfg_var.preference_for_host("github.com") == "https"

    cfg.get("git_remote_preferences").set("github.com", "ssh")
    assert git_remote_cfg_var.preference_for_host("github.com") == "ssh"

    cfg.get("git_remote_preferences").unset("gitlab.com")
    assert git_remote_cfg_var.preference_for_host("gitlab.com") == "https"

    cfg_dict_out = cfg.to_dict()
    cfg_dict_out_expected = {
        "git_remote_preferences": {
            "git.h2t.iar.kit.edu": "ssh",
            "github.com": "ssh"
        }
    }

    assert cfg_dict_out == cfg_dict_out_expected


def test_built_in_config():
    """Tests the overridden config with built-ins."""

    cfg = ConfigWithBuiltIns()

    assert cfg.get("dev_mode") is False
    cfg.set("dev_mode", True)
    assert cfg.get("dev_mode") is True
    cfg.unset("dev_mode")
    assert cfg.get("dev_mode") is False
    assert BuiltInConfigVariables.get_by_name("dev_mode").value is False


@pytest.fixture
def axii_config():
    """Fixture to setup a config."""

    config = AxiiConfig()
    yield config


def test_global_dev_mode(axii_config):
    """Test getting, setting and unsetting global variable dev_mode."""

    assert axii_config.get("dev_mode") is False
    assert "dev_mode" not in axii_config.global_config.config_vars
    axii_config.set("dev_mode", "enable", parse=True, is_global=True)
    assert axii_config.get("dev_mode") is True
    assert "dev_mode" in axii_config.global_config.config_vars
    axii_config.unset("dev_mode", is_global=True)
    assert axii_config.get("dev_mode") is False
    assert "dev_mode" not in axii_config.global_config.config_vars
